import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.util.*;


public class Ball extends Objekte
{
    public static final int typeCount = 8; // Es gibt acht Ballfarben

    // Hier ist die Liste der m?glichen B?lle
    private static GreenfootImage[] images = 
        {
            new GreenfootImage("bubble-1.png"),
            new GreenfootImage("bubble-2.png"),
            new GreenfootImage("bubble-3.png"),
            new GreenfootImage("bubble-4.png"),
            new GreenfootImage("bubble-5.png"),
            new GreenfootImage("bubble-6.png"),
            new GreenfootImage("bubble-7.png"),
            new GreenfootImage("bubble-8.png")
        };

    private static final double MOVE_SPEED = 6; // Geschwindigkeit wenn die B?lle von der Kanone geschossen werden
    private static final int FALL_HSPEED = 3; // horizontale Geschwindigkeit (random max)
    private static final int FALL_VSPEED = 10; // Anfangsgeschwindigkeit nach oben, wenn die B?lle "fallen"  (random max)
    private static final double FALL_GRAVITY = 0.12; // Eine Art Schwerkraft,Beschleunigung . speed every act().

    // The three states a ball can be in.
    public static final int STATE_MOVING = 0;    // Der Ball ist gerade am bewegen
    public static final int STATE_STUCK = 1;     // Der Ball steht
    public static final int STATE_FALLING = 2;   // Der Ball f?llt

    private int type; // Farbe des Balls
    private int state = STATE_STUCK; // momentaner Status des Balles
    public double exactX, exactY; // floating point position for smoothness.
    private double velocityX, velocityY; // floating point velocity.
    private PunkteStandAnzeige myScore;
    public Ball(int t)
    {
        type = t;
        
        // Hier wird die richtige Farbe gesetzt
        setImage(images[type]);
        
    }

    protected void addedToWorld(World world)
    {
        // Speichere die Position
        exactX = getX();
        exactY = getY();
    }

    public int getType()
    {
        return type;
    }

    public int getState()
    {
        return state;
    }

    // Der Ball befindet sich in einer Bewegung, desshalb wird der Winkel der Ballflugbahn berechnet
    public void move(double angle)
    {
        state = STATE_MOVING;
        double a = (angle + 270) * Math.PI / 180.0;
        velocityX = Math.cos(a) * MOVE_SPEED;
        velocityY = Math.sin(a) * MOVE_SPEED;
    }

    // Der Ball steht: Also keine Bewegung
    private void stick()
    {
        state = STATE_STUCK;
        
      
        velocityX = 0;
        velocityY = 0;
        
        // Hier soll ein freier Platz neben dem gesetzten Ball gefunden werden
        Map map = ((BubbleWorld)getWorld()).getMap();
        Map.Position pos = map.findFreeCell((int)exactX, (int)exactY);
        
        // Einsetzten der x,y Werte
        exactX = map.getX(pos.i, pos.j);
        exactY = map.getY(pos.i, pos.j);

        // Der Ball soll sich bei diesem freien Platz platzieren
        map.setBall(pos.i, pos.j, this);
        getWorld().addObject(new StickEffect(), (int)exactX, (int)exactY);
        
        // Der Ton zu Platzierung
        Greenfoot.playSound("stick.wav");
    }

    // WEnn er im Status Fallen ist soll der Ball fallen
    public void fall()
    {
        state = STATE_FALLING;

        // Anfangsgeschwindigkeit wird gesetzt
        velocityX += (Greenfoot.getRandomNumber(FALL_HSPEED)/10.0+0.5) * (Greenfoot.getRandomNumber(2)*2-1);
        velocityY -= (Greenfoot.getRandomNumber(FALL_VSPEED)/10.0+0.5);
    }

    private void checkCollisions()
    {
        int left = 16;
        int right = getWorld().getWidth()-16;

        if(exactX < left)
        {
            exactX -= velocityX*2;
            velocityX = -velocityX;
            
            // Ton
            Greenfoot.playSound("bounce.wav");
        }
        else if(exactX >= right)
        {
            exactX -= velocityX*2;
            velocityX = -velocityX;
            
            // Ton
            Greenfoot.playSound("bounce.wav");
        }

        if(exactY < 16)
        {
            stick();
        }
        else
        {
            List<Ball> others = getIntersectingObjects(Ball.class);
            double collisionDistance = Math.pow(32 * 0.82, 2);
            boolean tooClose = false;
            for(Ball b : others)
            {
                double dx = b.getX() - getX();
                double dy = b.getY() - getY();
                double d2 = dx*dx + dy*dy;

                if( b.getState() == Ball.STATE_STUCK &&
                    d2 <= collisionDistance)
                {
                    tooClose = true;
                }
            }

            if(tooClose)
            {
                stick();
            }

        }
    }
 


    public void act(double deltaTime) 
    {
        if(state == STATE_MOVING)
        {
            checkCollisions();

            exactX += velocityX * deltaTime;
            exactY += velocityY * deltaTime;
            setLocation((int)exactX, (int)exactY);   
        }
        else if(state == STATE_FALLING)
        {
            velocityY += FALL_GRAVITY * deltaTime;
            exactX += velocityX * deltaTime;
            exactY += velocityY * deltaTime;
            setLocation((int)exactX, (int)exactY);   

            if(exactY >= getWorld().getHeight()+getImage().getHeight()/2)
            {
                getWorld().removeObject(this);
            }
        }
        
       
       // if (exactY >= 500) {
        //((BubbleWorld) getWorld()).gameOver();
//}

    }
    
}
