import greenfoot.*;
import java.util.*;


public class Map  
{
    // Eine Hilfsklasse welche hilft die i und j positionen eines Objectes zu speichern 
    public class Position
    {
        public int i, j;
    }

    public static final int MAX_WIDTH = 22;       // Maximale Weite der Karte in Zellen 
    public static final int MAX_HEIGHT = 20;     // Maximale H?he der Karte in Zellen 
    public static final int COLUMN_WIDTH = 32;   // Platz zwischen den B?llen (in Pixel)
    public static final int ROW_HEIGHT = 28;     // Platz zwischen Ballreihen (in Pixel)

    private Cell[] cells = new Cell[MAX_WIDTH * MAX_HEIGHT]; // Die Karte selbst
    private int[] allowedBallTypes; // Array der gerade verfgbaren B?lle
    private boolean won; // setzten wenn die Karte gewonnen wurde
    private World world; // um Objekte in die Welt einfgen zu k?nnen
    public PunkteStandAnzeige PS;
    
    public Map(World w, String map)
    {
        world = w;
        PS = new PunkteStandAnzeige("Schuesse: 0");
        // Initiallisierung des Arrays
        for(int i = 0; i < cells.length; ++i)
        {
            cells[i] = new Cell(null, true);
        }

        // Karte wird geladen
        loadMap(map);
    }

    public void update()
    {
        // Schau auf die B?lle(ob sie fallen mssen)
        updateBalls();

        // Update um die zu verfgung stehenden B?llen
        // Die Kanone soll nun nur die zur verfgung stehenden B?lle aufnehmen k?nnen
        updateAllowedBallTypes();

    }

    public boolean hasWon()
    {
        return won;
    }

    public int[] getAllowedBallTypes()
    {
        return allowedBallTypes;
    }

    // Rckgabe der Ballreihen in vertikaler Richtung
    public int getCellCountY()
    {
        return MAX_HEIGHT;
    }

    // Rckgabe der Balreihen in horinzontaler Richtung
    public int getCellCountX(int j)
    {
        boolean odd = (j & 1) == 1;
        return odd ? MAX_WIDTH-1 : MAX_WIDTH;
    }

    //  Rckgabe der X-koordinaten positionen fr die gegebenen Hex-Koordinaten
    public int getX(int i, int j)
    {
        boolean odd = (j & 1) == 1;
        return 16 + i * COLUMN_WIDTH + (odd ? 16 : 0);
    }

    //  Rckgabe der Y-koordinaten positionen fr die gegebenen Hex-Koordinaten
    public int getY(int i, int j)
    {
        return 16 + j * ROW_HEIGHT;
    }

    // Lege einen Ball bei den gegebenen Hex Koordinaten ab
    // Zus?tzlich: Nachschauen ob B?lle in der nachbarschafft sind 
    public void setBall(int i, int j, Ball ball)
    {
        Cell c = getCell(i,j);

        if(c == null || c.getBall() != null)
        {
            return;
        }

        c.setBall(ball);

        handleMatches(i,j);
    }

    // Return the map entry for the given hex coordinates, or null.
    public Cell getCell(int i, int j)
    {
        if(i < 0 || j < 0 || i >= getCellCountX(j) || j >= getCellCountY())
        {
            return null;
        }

        return cells[i + j * MAX_WIDTH];
    }

    // Karte mit B?llen laden
    private void loadMap(String map)
    {
        // Lookup table for the string-based maps.
        String lookups = "0123456789 "; 
        PS = new PunkteStandAnzeige("Schuesse: 0");
        // Leert den Raum zwischen den B?llen
        map = map.replaceAll(" ","");

        // Schaut wie viele B?lle gew?hlt wurden
        int ci = 0;
        for(int j = 0; j < getCellCountY(); ++j)
        {
            for(int i = 0; i < getCellCountX(j); ++i)
            {
                // Stop if the map text ends.
                if(ci >= map.length())
                {
                    return;
                }

                // Translate the character in the map string to a ball type.
                int type = lookups.indexOf(map.charAt(ci++));

                // If it's a valid type, spawn a ball.
                if(type >= 0 && type < Ball.typeCount)
                {
                    // Ball einfgen
                    Ball ball = new Ball(type);
                    world.addObject(ball, getX(i,j), getY(i,j));


                    getCell(i,j).setBall(ball);
                }
            }
        }
    }

    // Given the position of a new ball, find matching ones and make them fall off the playfield.
    private void handleMatches(int i, int j)
    {
        // Generate a list of matching balls.
        ArrayList<Cell> matches = new ArrayList<Cell>();
        
        // Findet die zugeh?rigen B?lle
        match(i,j,matches);

        // Wir haben >= 3 Zugeh?rigkeiten!
        if(matches.size() >= 3)
        {
            for(Cell c : matches)
            {
                world.addObject(new StickEffect(), c.getBall().getX(), c.getBall().getY());
                c.getBall().fall();
                c.setBall(null);
            }
        }
    }

    // schaut auf die selben farben
    private void match(int i, int j, ArrayList<Cell> matches)
    {
        Cell c = getCell(i,j);

      
        // nicht ball einfgen dens nicht gibt
        if(c == null || c.getBall() == null || matches.contains(c))
        {
            return;
        }

      
        if(matches.isEmpty() || c.getBall().getType() == matches.get(0).getBall().getType())
        {
            // Aufnahme
            matches.add(c);

            // Nachbaren
            match(i-1, j, matches);
            match(i+1, j, matches);

            
            if((j & 1) == 1) // odd row
            {
                match(i, j-1, matches);
                match(i+1, j-1, matches);
                match(i, j+1, matches);
                match(i+1, j+1, matches);
            }
            else // even row
            {
                match(i-1, j-1, matches);
                match(i, j-1, matches);
                match(i-1, j+1, matches);
                match(i, j+1, matches);
            }
        }
    }

  
    public void updateAllowedBallTypes()
    {
        int allowedCount = 0;
        boolean[] allowed = new boolean[Ball.typeCount];
        
        // Nur Baelle der selben farbe werden genehmigt
        for(Cell c : cells)
        {
            if(c != null && c.getBall() != null && c.isAttached())
            {
                int type = c.getBall().getType();
                
                if(!allowed[type])
                    allowedCount++;
                
                allowed[type] = true;
            }
        }
        
        allowedBallTypes = new int[allowedCount];
        int writeIndex = 0;
        for(int type = 0; type < Ball.typeCount; ++type)
        {
            if(allowed[type])
            {
                allowedBallTypes[writeIndex++] = type;
            }
        }
    }

    private void updateBalls()
    {
        
        for(Cell c : cells)
        {
            c.setAttached(false);
        }

        // Find the balls attached to the top edge of the map, and scan ALL their neighbors recursively.
       
        for(int i = 0; i < getCellCountX(0); ++i)
        {
            Cell c = cells[i];

            if(c.getBall() != null)
            {
                markAttached(i,0);
            }
        }

        // Clear any balls that don't have the attached flag set.
        // ALSO set the 'won' attribute if the game was won (== no attached balls on the map).
        won = true;
        for(Cell c : cells)
        {
            if(c != null && c.getBall() != null)
            {
                if(c.isAttached())
                {
                    won = false;
                }
                else
                {
                    c.getBall().fall();
                    c.setBall(null);
                }
            }
        }
    }

    private void markAttached(int i, int j)
    {
        Cell c = getCell(i,j);

        // Don't continue with:
        // 1) Invalid cells 
        // 2) Empty cells
        // 3) Already attached cells
        if(c == null || c.getBall() == null || c.isAttached())
        {
            return;
        }

        c.setAttached(true);

        // Left/right neighbors.
        markAttached(i-1, j);
        markAttached(i+1, j);

        // Odd and even rows have their diagonal neighbors positioned differently in i,j coordinates.
        if((j & 1) == 1) // odd row
        {
            markAttached(i, j-1);
            markAttached(i+1, j-1);
            markAttached(i, j+1);
            markAttached(i+1, j+1);
        }
        else // even row
        {
            markAttached(i-1, j-1);
            markAttached(i, j-1);
            markAttached(i-1, j+1);
            markAttached(i, j+1);
        }
    }

    // Return the nearest free space to the specified pixel coordinates.
    public Position findFreeCell(int x, int y)
    {
        // Linear search through the map, just return the closest map entry position.
        // This could be more efficient.
        // Durchsicht die Welt, 
        double minDistance = Double.MAX_VALUE;
        Position pos = null;

        for(int j = 0; j < getCellCountY(); ++j)
        {
            for(int i = 0; i < getCellCountX(j); ++i)
            {
                Cell c = getCell(i,j);

                // Only consider empty cells.
                if(c != null && c.getBall() == null)
                {
                    double dx = getX(i,j) - x;
                    double dy = getY(i,j) - y;
                    double distance = dx*dx+dy*dy;

                    if(distance < minDistance)
                    {
                        minDistance = distance;

                        if(pos == null)
                        {
                            pos = new Position();
                        }

                        pos.i = i;
                        pos.j = j;
                    }
                }
            }
        }

        // Return the position (or null if none found)
        return pos;
    }

}
